from __future__ import absolute_import

import threading
import select
import json
try:
    from queue import Queue
except ImportError:
    from Queue import Queue

from websocket import WebSocket as BaseWebSocket


class Replies(object):
    def __init__(self):
        self.replies = {}

    def __getitem__(self, reply_id):
        reply = self.replies[reply_id]
        del self.replies[reply_id]
        return reply

    def __setitem__(self, reply_id, message):
        self.replies[reply_id] = message


class WebSocket(BaseWebSocket):
    def __init__(self, on_event, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)

        self._on_event = on_event

        self.request_id = 1
        self.cLock = threading.Condition()
        self.replies = Replies()
        self.events = Queue()

    def connect(self, *args, **kwargs):
        super(self.__class__, self).connect(*args, **kwargs)

        threading.Thread(target=self._read_messages).start()
        threading.Thread(target=self._process_events).start()

    def close(self):
        # Terminates blocking `get` call on events processing thread
        self.events.put(None)
        return super(self.__class__, self).close()

    def send_message(self, method, params):
        request_id = self.request_id
        self.request_id += 1

        payload = json.dumps({
            'id': request_id,
            'method': method,
            'params': params,
        })
        self.send(payload)

        # Wait until response is ready
        with self.cLock:
            self.cLock.wait()

        return self.replies[request_id]

    def _read_messages(self):
        while self.connected:
            r, w, e = select.select([self.sock], [], [])

            if r and self.connected:
                with self.cLock:
                    data = self.recv()
                    message = json.loads(data)

                    # If id is present -> notifies main thread to return this message as a reply
                    # Otherwise puts message into events queue so it will be dispatched as event
                    if 'id' in message:
                        self.replies[message['id']] = message
                        self.cLock.notify()
                    else:
                        self.events.put(message)

    def _process_events(self):
        while self.connected:
            event = self.events.get()

            # `None` is standing as a null terminator event
            if event is None:
                break

            self._on_event(event)



class BaseDomain(object):
    pass


def get_command(domain_name, command_name):
    """Returns a closure function that dispatches message to the WebSocket."""
    def send_command(self, **kwargs):
        return self.ws.send_message(
            '{0}.{1}'.format(domain_name, command_name),
            kwargs
        )

    return send_command


def DomainFactory(domain_name, cmds):
    """Dynamically create Domain class and set it's methods."""
    klass = type(str(domain_name), (BaseDomain,), {})

    for c in cmds:
        command = get_command(domain_name, c['name'])
        setattr(klass, c['name'], classmethod(command))

    return klass



from __future__ import absolute_import

import json
from os import path
try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen

from cproto.core.websocket import WebSocket
from cproto.domains.factory import DomainFactory

ROOT_DIR = path.abspath(path.dirname(path.dirname(__file__)))


class CProto(object):

    def __init__(self, host='127.0.0.1', port=9222):
        res = urlopen('http://{0}:{1}/json'.format(host, port))

        url = json.loads(res.read())[0]['webSocketDebuggerUrl']
        self.ws = WebSocket(on_event=self._on_event)
        self.ws.connect(url)

        with open(path.join(ROOT_DIR, 'resources/protocol.json'), 'rb') as f:
            data = json.loads(f.read())

        for d in data['domains']:
            domain_name = d['domain']

            # Build Domain Class from protocol
            DomainClass = DomainFactory(domain_name, d['commands'])

            # Set WebSocket attribute
            DomainClass.ws = self.ws

            # Set Domain's Class as a property
            setattr(self, domain_name, DomainClass())

    def close(self):
        self.ws.close()

    def _on_event(self, message):
        domain_name, method_name = message['method'].split('.')
        domain = getattr(self, domain_name)

        if hasattr(domain, method_name):
            getattr(domain, method_name)(message['params'])
